#include "cEnvelopeComponent.h"
#include "resources.h"

#include <algorithm>

///////////////////////////////////////////////////////////////////////////////////////////////////
// cEnvelopeAnchor
///////////////////////////////////////////////////////////////////////////////////////////////////
cEnvelopeAnchor::cEnvelopeAnchor(void)
    : mCurX(0),
      mCurY(0),
      mMaxX(0),
      mMaxY(0),
      mEnabled(true),
      mActive(false),
      mSelected(false),
      mAnchorToLeft(nullptr),
      mAnchorToRight(nullptr),
      mColor(Colours::white),
      mSize(7)
{
    UpdateDisplayState();
}

void cEnvelopeAnchor::Display( Graphics& g, int xOffset, int yOffset )
{
    // draw an anchor if the anchor can move at least in one direction
    if ((!mX.mLocked) || (!mY.mLocked))
    {
        drawCircle(g, (float)(mCurX + xOffset), (float)(mCurY + yOffset), 1.0, 8.0, mColor);
    }
}

void cEnvelopeAnchor::drawCircle(Graphics& g, float centerX, float centerY, float startSize, float size, Colour color)
{
    float lineWidth = 1.0f;
    float alpha = 1.0f;
    float endSize = startSize + size;
    float alphaStep = 1.0f / (size / lineWidth);
    for (float curSize = startSize; curSize < endSize; curSize += lineWidth)
    {
        float curRadius = curSize / 2.0f;
        g.setColour(color.withAlpha(alpha));
        g.drawEllipse(Rectangle<float>(centerX - curRadius, centerY - curRadius, curSize, curSize), lineWidth);
        alpha -= alphaStep;
    }
//     float radius = endSize / 2.0f;
//      g.setColour(color.withAlpha(1.0f));
//      g.drawEllipse(Rectangle<float>(centerX - radius, centerY - radius, endSize, endSize), lineWidth);
}

void cEnvelopeAnchor::UpdateDisplayState( void )
{
    mColor = Colours::cyan;
    if ((!mX.mLocked) || (!mY.mLocked))
        if (!mEnabled)
            mColor = Colours::darkgrey;
        else if (mActive)
            if (mSelected)
                mColor = Colours::yellow;
            else
                mColor = Colours::white;
        else 
           if ( mSelected )
               mColor = Colours::greenyellow;
           else
               mColor = Colours::cyan;
}   

int cEnvelopeAnchor::DistanceMovedX(int xMove)
{
    int distanceMoved = xMove;

    if (mX.mLocked)
    {
        distanceMoved = 0;
    }
    else if (xMove < 0)      // moving to the left
    {
        // if this is the furthest left anchor, and the move would be to less than zero, clamp the value to -mCurX
        if (mAnchorToLeft == kNoAnchor)
        {
            if (mCurX + xMove < 0)
                distanceMoved = -mCurX;
        }
        else if (mAnchorToLeft->mX.mLocked)    // anchor to left is locked, can only go as far as anchor to lefts X
        {
            distanceMoved = mAnchorToLeft->mCurX - mCurX;
        }
        else if (!mAnchorToLeft->IsSelected())  // anchor to left is NOT selected
        {
            // not selected, can only go as far as anchor to lefts X
            {
                if (mCurX + xMove < mAnchorToLeft->mCurX)
                    distanceMoved = mAnchorToLeft->mCurX - mCurX;

            }
        }
        else    // otherwise, it IS selected, so let's see how far it can move
        {
            distanceMoved = mAnchorToLeft->DistanceMovedX(distanceMoved);
        }
    }
    else if (xMove > 0) // moving to the right
    {
        // if this is the furthest right anchor, and the move would be to greater than mMaxX, clamp the value to -mCurX
        if (mAnchorToRight == kNoAnchor)
        {
            if (mCurX + xMove > mMaxX)
                distanceMoved = mMaxX - mCurX;
        }
        else if (mAnchorToRight->mX.mLocked)    // anchor to right is locked, can only go as far as anchor to rights X
        {
            distanceMoved = mAnchorToRight->mCurX - mCurX;
        }
        else if (!mAnchorToRight->IsSelected())
        {
            // not selected, can only go as far as anchor to rights X
            {
                if (mCurX + xMove > mAnchorToRight->mCurX)
                    distanceMoved = mAnchorToRight->mCurX - mCurX;
            }
        }
        else
        {
            distanceMoved = mAnchorToRight->DistanceMovedX(distanceMoved);
        }
    }

    return distanceMoved;
}

int cEnvelopeAnchor::DistanceMovedY(int yMove)
{
    int distanceMoved = yMove;

    if (mY.mLocked)
    {
        distanceMoved = 0;
    }
    if (yMove < 0)      // moving up
    {
        if ((mCurY + yMove) < 0)
            distanceMoved = -mCurY;
    }
    else if (yMove > 0) // moving down
    {
        if ((mCurY + yMove) > mMaxY)
            distanceMoved = mMaxY - mCurY;
    }

    return distanceMoved;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// cEnvelopeComponent
///////////////////////////////////////////////////////////////////////////////////////////////////
cEnvelopeComponent::cEnvelopeComponent()
    : mCurActiveAnchor( kNoAnchor ),
      mDraggingAnchor( kNoAnchor ),
      mEnvelopeDisplayAreaHeight( 0 ),
      mEnvelopeDisplayAreaWidth( 0 ),
      mOffset(0),
      mBorderWidth( kDefaultBorderWidth ),
      mAnchorSize( kDefaultHandleSize ),
      mBackgroundColour( kDefaultBackgroundColour ),
      mBorderColour( kDefaultBorderColour ),
      mLineColour( kDefaultLineColour ),
      mGlowColour(kDefaultGlowColour),
      mNormalHandleColour(kDefaultNormalHandleColour),
      mActiveHandleColour( kDefaultActiveHandleColour ),
      mDisabledHandleColour( kDefaultDisabledHandleColour ),
      mStartX(0),
      mStartY(0),
      mStartRectSelectX(-1),
      mStartRectSelectY(-1),
      mCurRectSelectX(-1),
      mCurRectSelectY(-1)
{
    mBackgroundImage = ImageCache::getFromMemory(resources::background_png, resources::background_pngSize);

    mOffset = mAnchorSize / 2;
}

cEnvelopeComponent::~cEnvelopeComponent()
{
}

/////////////////////
// Anchor management
/////////////////////

// inserts an anchor at the position indicated by the index, with -1 indicating the end (ie. an append)
void cEnvelopeComponent::addAnchor(cEnvelopeAnchor* anchor, int index)
{
    anchor->SetMaxX(mEnvelopeDisplayAreaWidth);
    anchor->SetMaxY(mEnvelopeDisplayAreaHeight);

    mEnvelopeAnchors.insert(index, anchor);

    if (index == 0)
    {
        // this was a prepend
        if (index < (mEnvelopeAnchors.size() - 1))
        {
            mEnvelopeAnchors[ index ]->mAnchorToRight = mEnvelopeAnchors[ index + 1 ];
            mEnvelopeAnchors[ index  + 1 ]->mAnchorToLeft = mEnvelopeAnchors[ index ];
        }
    }
    else if ((index < 0) || (index > (mEnvelopeAnchors.size() - 2)))
    {
        // this was an append
        index = mEnvelopeAnchors.size() - 1;

        if (index > 0)
        {
            mEnvelopeAnchors[ index ]->mAnchorToLeft = mEnvelopeAnchors[ index - 1 ];
            mEnvelopeAnchors[ index  - 1 ]->mAnchorToRight = mEnvelopeAnchors[ index ];
        }
    }
    else
    {
        // this is an insert
        if (index > 0)
        {
            mEnvelopeAnchors[ index ]->mAnchorToLeft = mEnvelopeAnchors[ index - 1 ];
            mEnvelopeAnchors[ index  - 1 ]->mAnchorToRight = mEnvelopeAnchors[ index ];
        }
        if (index < (mEnvelopeAnchors.size() - 1))
        {
            mEnvelopeAnchors[ index ]->mAnchorToRight = mEnvelopeAnchors[ index + 1 ];
            mEnvelopeAnchors[ index  + 1 ]->mAnchorToLeft = mEnvelopeAnchors[ index ];
        }
    }

    repaint();
}

// deletes an anchor at the position indicated by the index, causing all anchors to the right of the index to shift one left
cEnvelopeAnchor* cEnvelopeComponent::deleteAnchor(int index)
{
    cEnvelopeAnchor* anchorThatWasDeleted = nullptr;
    if ((index >= 0) || (index < mEnvelopeAnchors.size()))
    {
        anchorThatWasDeleted = mEnvelopeAnchors[ index ];

        mEnvelopeAnchors.remove(index);

        if (index == 0)
        {
            // deleted the first anchor
            if (mEnvelopeAnchors.size() > 0)
                mEnvelopeAnchors[ 0 ]->mAnchorToLeft = kNoAnchor;
        }
        else if (index == mEnvelopeAnchors.size() )
        {
            // deleted the last anchor
            mEnvelopeAnchors[ mEnvelopeAnchors.size() - 1 ]->mAnchorToRight = kNoAnchor;
        }
        else
        {
            mEnvelopeAnchors[ index ]->mAnchorToLeft = mEnvelopeAnchors[ index - 1 ];
            mEnvelopeAnchors[ index  - 1 ]->mAnchorToRight = mEnvelopeAnchors[ index ];
        }
    }

    repaint();

    return anchorThatWasDeleted;
}

void cEnvelopeComponent::initAnchors( long numberOfEnvelopeAnchors )
{
    for( int curEnvelopAnchor = 0; curEnvelopAnchor < numberOfEnvelopeAnchors; ++curEnvelopAnchor )
        addAnchor(new cEnvelopeAnchor);
}

void cEnvelopeComponent::setAnchorXPercent( long anchorIndex, double xPercent )
{
    if( anchorIndex < mEnvelopeAnchors.size() )
    {
        mEnvelopeAnchors[ anchorIndex ]->mX.mPercentage = xPercent;
        mEnvelopeAnchors[anchorIndex]->recalculateX(mBorderWidth, mEnvelopeDisplayAreaWidth);
    }
}

void cEnvelopeComponent::setAnchorYPercent( long anchorIndex, double yPercent )
{
    if( anchorIndex < mEnvelopeAnchors.size() )
    {
        mEnvelopeAnchors[ anchorIndex ]->mY.mPercentage = yPercent;
        mEnvelopeAnchors[anchorIndex]->recalculateY(mBorderWidth, mEnvelopeDisplayAreaHeight);
    }
}

double cEnvelopeComponent::getAnchorXPercent( long anchorIndex )
{
    double anchorXPercent = 0.0;

    if( anchorIndex < mEnvelopeAnchors.size() )
    {
        anchorXPercent = mEnvelopeAnchors[ anchorIndex ]->mX.mPercentage;
    }

    return( anchorXPercent );
}

double cEnvelopeComponent::getAnchorYPercent( long anchorIndex )
{
    double anchorYPercent = 0.0;

    if( anchorIndex < mEnvelopeAnchors.size() )
    {
        anchorYPercent = mEnvelopeAnchors[ anchorIndex ]->mY.mPercentage;
    }

    return( anchorYPercent );
}

// since anchor positions are stored in 'percentages', we need to recalculate x/y positions if either the position of an anchor changes, or the dimensions of the containing window changes
void cEnvelopeComponent::recalculateCoordinates( void )
{
    for( long curAnchor = 0; curAnchor < mEnvelopeAnchors.size(); ++curAnchor )
    {
        mEnvelopeAnchors[curAnchor]->recalculateX(mBorderWidth, mEnvelopeDisplayAreaWidth);
        mEnvelopeAnchors[curAnchor]->recalculateY(mBorderWidth, mEnvelopeDisplayAreaHeight);
    }
}

//////////////////////////////////////////////////////////////////////////////////
//
// Mouse Behaviors
//
//////////////////////////////////////////////////////////////////////////////////
void cEnvelopeComponent::rightMouseClickOnAnchor( const MouseEvent &/*e*/ )
{
    cEnvelopeAnchor* selectedAnchor = mCurActiveAnchor;

    mCurActiveAnchor = kNoAnchor;
    mDraggingAnchor = kNoAnchor;

    PopupMenu m;
    m.addItem (1, "Properties...");
    m.addItem (2, "Delete...");

    const int result = m.show();

    if (result == 0)
    {
        // user dismissed the menu without picking anything
    }
    else if (result == 1)
    {
        // user picked item 1
    }
    else if (result == 2)
    {
        for (int curEnvelopeAnchorIndex = 0; curEnvelopeAnchorIndex < mEnvelopeAnchors.size(); ++curEnvelopeAnchorIndex)
        {
            if (mEnvelopeAnchors[ curEnvelopeAnchorIndex ] == selectedAnchor)
            {
                deleteAnchor(curEnvelopeAnchorIndex);
                selectedAnchor = kNoAnchor;
                break;
            }
        }
    }

    if (selectedAnchor != kNoAnchor)
        selectedAnchor->SetActive(false);
}

void cEnvelopeComponent::rightMouseClickOnBackground( const MouseEvent &e )
{
    // if the ctrl key is being held down, then this is a request to add
    if (e.mods.isCtrlDown())
    {
        int colClicked = e.x;

        // verify that the click was after the first anchor, and before the last anchor
        if ((colClicked > mEnvelopeAnchors[ 0 ]->mCurX) && (colClicked < mEnvelopeAnchors[ mEnvelopeAnchors.size() - 1 ]->mCurX))
        {
            // find out at which index we should add this new anchor
            for (int curEnvelopeAnchorIndex = 1; curEnvelopeAnchorIndex < mEnvelopeAnchors.size() - 1; ++curEnvelopeAnchorIndex)
            {
                if (colClicked < mEnvelopeAnchors[ curEnvelopeAnchorIndex ]->mCurX )
                {
                    cEnvelopeAnchor* newEnvelopeAnchor = new cEnvelopeAnchor;
                    newEnvelopeAnchor->mCurX = colClicked;
                    newEnvelopeAnchor->mCurY = mEnvelopeAnchors[ curEnvelopeAnchorIndex ]->mCurY;
                    newEnvelopeAnchor->mX.mPercentage = (float)e.x / (float)mEnvelopeDisplayAreaWidth;
                    newEnvelopeAnchor->recalculateX(mBorderWidth, mEnvelopeDisplayAreaWidth);
                    newEnvelopeAnchor->mY.mPercentage = mEnvelopeAnchors[ curEnvelopeAnchorIndex ]->mY.mPercentage;
                    newEnvelopeAnchor->recalculateY(mBorderWidth, mEnvelopeDisplayAreaHeight);
                    addAnchor(newEnvelopeAnchor, curEnvelopeAnchorIndex );
                    break;
                }
            }
        }
    }
    else
    {
        //////////////////////////
        // test code
        // test code toggle enabled state
        setEnabled(!isEnabled());
        //////////////////////////
    }
}

// called when the mouse is clicked on an anchor
void cEnvelopeComponent::mouseDownOnAnchor(const MouseEvent &e)
{
    // shift key - add to current selection
    // ctrl key  - toggle current selection
    if (isEnabled())
    {
        mResetWhenDone = false;
        if (e.mods.isRightButtonDown())
        {
            rightMouseClickOnAnchor(e);
        }
        else
        {
            // check modifier keys to determine selection type
            if (e.mods.isShiftDown())
            {
                // if shift is held down, setting anchor selection

                // set current as selected
                mCurActiveAnchor->SetSelected(true);
            }
            else if (e.mods.isCtrlDown())
            {
                // if ctrl is held down, toggling anchor selection

                // toggle the current selection setting
                mCurActiveAnchor->SetSelected(!mCurActiveAnchor->IsSelected());
            }
            else
            {
                // no modifier keys held down, this is an individual anchor selection

                // if this anchor isn't selected, then let's make sure none of the other anchors are selected
                // otherwise, if it is selected, we are possibly starting a move operation
                if (!mCurActiveAnchor->IsSelected())
                {
                    for (int curAnchorIndex = 0; curAnchorIndex < mEnvelopeAnchors.size(); ++curAnchorIndex)
                        mEnvelopeAnchors[curAnchorIndex]->SetSelected(false);
                }

                if (!mCurActiveAnchor->IsSelected())
                    mResetWhenDone = true;
                mCurActiveAnchor->SetSelected(true);
            }

            if (mCurActiveAnchor->IsSelected())
            {
                mDraggingAnchor = mCurActiveAnchor;
                mStartX = mDraggingAnchor->mCurX;
                mStartY = mDraggingAnchor->mCurY;
            }
        }
        repaint();
    }
}

// called when mouse is clicked on the background
void cEnvelopeComponent::mouseDownOnBackground(const MouseEvent &e)
{
    if (e.mods.isRightButtonDown())
    {
        rightMouseClickOnBackground(e);
    }
    else
    {
        // start rectangle selection
        mStartRectSelectX = e.getPosition().getX();
        mStartRectSelectY = e.getPosition().getY();
        mCurRectSelectX = mStartRectSelectX;
        mCurRectSelectY = mStartRectSelectY;
    }

    repaint();
}

// called when done dragging an anchor
void cEnvelopeComponent::mouseUpDraggingAnchor(const MouseEvent &e)
{
    // no modifier keys pressed, 
    if (!e.mods.isShiftDown() && !e.mods.isCtrlDown())
    {
        // if mouse didn't move, clear all the selected ones
        if ((mStartX == mCurActiveAnchor->mCurX) && (mStartY == mCurActiveAnchor->mCurY))
        {
            for (int curAnchorIndex = 0; curAnchorIndex < mEnvelopeAnchors.size(); ++curAnchorIndex)
            {
                mEnvelopeAnchors[curAnchorIndex]->SetSelected(false);
            }
        }
        else
        {
            // if mouse did move
            if (mResetWhenDone)
            {
                bool anotherSelected = false;
                for (int curAnchorIndex = 0; curAnchorIndex < mEnvelopeAnchors.size(); ++curAnchorIndex)
                {
                    if ((mCurActiveAnchor != mEnvelopeAnchors[curAnchorIndex]) && (mEnvelopeAnchors[curAnchorIndex]->IsSelected()))
                        anotherSelected = true;
                }
                // if there weren't any other anchors selected, reset this one, otherwise leave everything selected
                if (!anotherSelected)
                    mCurActiveAnchor->SetSelected(false);

                mResetWhenDone = false;
            }
        }

        repaint();
    }

    mDraggingAnchor = kNoAnchor;
}

// called when done creating a rectangle selection
void cEnvelopeComponent::mouseUpRectangleSelection(const MouseEvent &e)
{
    Logger::outputDebugString("-- doing rectangular selection --");
    int width = mCurRectSelectX - mStartRectSelectX;
    int height = mCurRectSelectY - mStartRectSelectY;
    int startSelectX = mStartRectSelectX;
    int startSelectY = mStartRectSelectY;
    if (width < 0)
    {
        startSelectX += width;
        width = abs(width);
    }
    if (height < 0)
    {
        startSelectY += height;
        height = abs(height);
    }
    Rectangle<float> selectionRect((float)startSelectX, (float)startSelectY, (float)width, (float)height);
    for (int curAnchorIndex = 0; curAnchorIndex < mEnvelopeAnchors.size(); ++curAnchorIndex)
    {
        if (selectionRect.contains((float)mEnvelopeAnchors[curAnchorIndex]->mCurX, (float)mEnvelopeAnchors[curAnchorIndex]->mCurY))
        {
            // this anchor is within the selection rectangle

            // if ctrl is held, but not shift, toggle the anchor setting
            if (e.mods.isCtrlDown() && !e.mods.isShiftDown())
            {
                mEnvelopeAnchors[curAnchorIndex]->SetSelected(!mEnvelopeAnchors[curAnchorIndex]->IsSelected());
            }
            else
            {
                // set anchor
                mEnvelopeAnchors[curAnchorIndex]->SetSelected(true);
            }
        }
        else
        {
            // this anchor is not within the selection rectangle

            // if shift is not being held, deselect any anchors outside of selection rectangle
            if (!e.mods.isShiftDown() && !e.mods.isCtrlDown())
            {
                mEnvelopeAnchors[curAnchorIndex]->SetSelected(false);
            }
        }
    }

    // reset the rectangle selection settings
    mStartRectSelectX = -1;
    mStartRectSelectY = -1;
    mCurRectSelectX = -1;
    mCurRectSelectY = -1;
    repaint();
}

// called when dragging an anchor
void cEnvelopeComponent::mouseDragAnchor(const MouseEvent &e)
{
    bool changed = false;

    if (!mDraggingAnchor->mX.mLocked)
    {
        long minX;
        long maxX;

        cEnvelopeAnchor* curAnchor = mDraggingAnchor;
        if (curAnchor->mAnchorToLeft == kNoAnchor)
        {
            minX = mBorderWidth;
        }
        else
        {
            minX = curAnchor->mAnchorToLeft->mCurX;
            curAnchor = curAnchor->mAnchorToLeft;

            while (curAnchor->mAnchorToLeft != kNoAnchor && curAnchor->IsSelected())
            {
                minX = curAnchor->mAnchorToLeft->mCurX;
                curAnchor = curAnchor->mAnchorToLeft;
            }
            if (curAnchor->mAnchorToLeft == kNoAnchor)
            {
                minX = mBorderWidth;
            }
        }

        curAnchor = mDraggingAnchor;
        if (curAnchor->mAnchorToRight == kNoAnchor)
        {
            maxX = mBorderWidth + mEnvelopeDisplayAreaWidth;
        }
        else
        {
            maxX = curAnchor->mAnchorToRight->mCurX;
            curAnchor = curAnchor->mAnchorToRight;

            while (curAnchor->mAnchorToRight != kNoAnchor && curAnchor->IsSelected())
            {
                maxX = curAnchor->mAnchorToRight->mCurX;
                curAnchor = curAnchor->mAnchorToRight;
            }
            if (curAnchor->mAnchorToRight == kNoAnchor)
            {
                maxX = mBorderWidth + mEnvelopeDisplayAreaWidth;
            }
        }

        long newX = std::min(std::max((long)e.x, minX), maxX);

        if (newX != mDraggingAnchor->mCurX)
        {
            long xOffset = newX - mDraggingAnchor->mCurX;

            for (int curAnchorIndex = 0; curAnchorIndex < mEnvelopeAnchors.size(); ++curAnchorIndex)
            {
                if (mEnvelopeAnchors[curAnchorIndex]->IsSelected())
                {
                    int newMaxXOffset = mEnvelopeAnchors[curAnchorIndex]->DistanceMovedX(xOffset);

                    if ((xOffset < 0) && (newMaxXOffset > xOffset))
                        xOffset = newMaxXOffset;
                    else if ((xOffset > 0) && (newMaxXOffset < xOffset))
                        xOffset = newMaxXOffset;
                }
            }

            // update anchor being dragged, both pixel value and percent value
            if (xOffset < 0)
            {
                for (int curAnchorIndex = 0; curAnchorIndex < mEnvelopeAnchors.size(); ++curAnchorIndex)
                {
                    if (mEnvelopeAnchors[curAnchorIndex]->IsSelected())
                    {
                        mEnvelopeAnchors[curAnchorIndex]->mCurX += xOffset;
                        mEnvelopeAnchors[curAnchorIndex]->recalculateXPercent(mBorderWidth, mEnvelopeDisplayAreaWidth);

                        cEnvelopeAnchor* anchorToLeft = mEnvelopeAnchors[curAnchorIndex]->mAnchorToLeft;
                        while ((anchorToLeft != NULL) && (anchorToLeft->mCurX > mEnvelopeAnchors[curAnchorIndex]->mCurX))
                        {
                            anchorToLeft->mCurX = mEnvelopeAnchors[curAnchorIndex]->mCurX;
                            anchorToLeft->recalculateXPercent(mBorderWidth, mEnvelopeDisplayAreaWidth);

                            anchorToLeft = anchorToLeft->mAnchorToLeft;
                        }

                    }
                }
            }
            else
            {
                for (int curAnchorIndex = mEnvelopeAnchors.size() - 1; curAnchorIndex >= 0; --curAnchorIndex)
                {
                    if (mEnvelopeAnchors[curAnchorIndex]->IsSelected())
                    {
                        mEnvelopeAnchors[curAnchorIndex]->mCurX += xOffset;
                        mEnvelopeAnchors[curAnchorIndex]->recalculateXPercent(mBorderWidth, mEnvelopeDisplayAreaHeight);

                        cEnvelopeAnchor* anchorToRight = mEnvelopeAnchors[curAnchorIndex]->mAnchorToRight;
                        while ((anchorToRight != NULL) && (anchorToRight->mCurX < mEnvelopeAnchors[curAnchorIndex]->mCurX))
                        {
                            anchorToRight->mCurX = mEnvelopeAnchors[curAnchorIndex]->mCurX;
                            anchorToRight->recalculateXPercent(mBorderWidth, mEnvelopeDisplayAreaHeight);

                            anchorToRight = anchorToRight->mAnchorToRight;
                        }
                    }
                }
            }

            // update anchor's that have 'mFollowLeft' set, both pixel value and percent value
            cEnvelopeAnchor* curAnchor = mDraggingAnchor->mAnchorToLeft;
            while ((curAnchor != NULL) && (curAnchor->mX.mFollowLeft))
            {
                // exclude any selected anchors, as they have already been moved by the previous code
                if (!curAnchor->IsSelected())
                {
                    curAnchor->mCurX += xOffset;
                    curAnchor->recalculateXPercent(mBorderWidth, mEnvelopeDisplayAreaWidth);
                }

                curAnchor = curAnchor->mAnchorToLeft;
            }

            // update anchor's that have 'mFollowRight' set, both pixel value and percent value
            curAnchor = mDraggingAnchor->mAnchorToRight;
            while ((curAnchor != NULL) && (curAnchor->mX.mFollowRight))
            {
                if (!curAnchor->IsSelected())
                {
                    curAnchor->mCurX += xOffset;
                    curAnchor->recalculateXPercent(mBorderWidth, mEnvelopeDisplayAreaWidth);
                }
                curAnchor = curAnchor->mAnchorToRight;
            }

            changed = true;
        }
    }

    if (!mDraggingAnchor->mY.mLocked)
    {
        long newY = std::min(std::max((long)e.y, (long)mBorderWidth), (long)(getHeight() - mBorderWidth));

        if (newY != mDraggingAnchor->mCurY)
        {
            long yOffset = newY - mDraggingAnchor->mCurY;

            for (int curAnchorIndex = 0; curAnchorIndex < mEnvelopeAnchors.size(); ++curAnchorIndex)
            {
                if (mEnvelopeAnchors[curAnchorIndex]->IsSelected())
                {
                    int newMaxYOffset = mEnvelopeAnchors[curAnchorIndex]->DistanceMovedY(yOffset);

                    if ((yOffset < 0) && (newMaxYOffset > yOffset))
                        yOffset = newMaxYOffset;
                    else if ((yOffset > 0) && (newMaxYOffset < yOffset))
                        yOffset = newMaxYOffset;
                }
            }

            // update anchor being dragged, both pixel value and percent value
            for (int curAnchorIndex = 0; curAnchorIndex < mEnvelopeAnchors.size(); ++curAnchorIndex)
            {
                if (mEnvelopeAnchors[curAnchorIndex]->IsSelected())
                {
                    mEnvelopeAnchors[curAnchorIndex]->mCurY += yOffset;
                    mEnvelopeAnchors[curAnchorIndex]->recalculateYPercent(mBorderWidth, mEnvelopeDisplayAreaHeight);
                }
            }

            if (mDraggingAnchor->mY.mFollowLeft)
            {
                cEnvelopeAnchor* leftAnchor = mDraggingAnchor;

                do
                {
                    leftAnchor = leftAnchor->mAnchorToLeft;

                    if (!leftAnchor->IsSelected())
                    {
                        leftAnchor->mCurY += yOffset;
                        leftAnchor->recalculateYPercent(mBorderWidth, mEnvelopeDisplayAreaHeight);
                    }

                } while (leftAnchor->mY.mFollowLeft);
            }

            if (mDraggingAnchor->mY.mFollowRight)
            {
                cEnvelopeAnchor* rightAnchor = mDraggingAnchor;

                do
                {
                    rightAnchor = rightAnchor->mAnchorToRight;

                    if (!rightAnchor->IsSelected())
                    {
                        rightAnchor->mCurY += yOffset;
                        rightAnchor->recalculateYPercent(mBorderWidth, mEnvelopeDisplayAreaHeight);
                    }

                } while (rightAnchor->mY.mFollowRight);
            }

            changed = true;
        }
    }

    if (changed)
    {
        // inform any users of us that something has changed
        sendChangeMessage();

        repaint();
    }
}

// called when dragging/creating a rectangle selection
void cEnvelopeComponent::mouseDragRectangleSelection(const MouseEvent &e)
{
    mCurRectSelectX = e.getPosition().getX();
    mCurRectSelectY = e.getPosition().getY();

    repaint();
}

////////////////////////////
// JUCE base class overrides
////////////////////////////
void cEnvelopeComponent::mouseDown(const MouseEvent &e)
{
    // if the mouse is currently over an anchor
    if (mCurActiveAnchor != kNoAnchor)
    {
        mouseDownOnAnchor(e);
    }
    else
    {
        // handle a mouse down that is not on a anchor
        mouseDownOnBackground(e);
    }
}

void cEnvelopeComponent::mouseUp(const MouseEvent &e)
{
    // if we are currently dragging an anchor
    if (mDraggingAnchor != kNoAnchor)
    {
        mouseUpDraggingAnchor(e);
    }
    else if (mStartRectSelectX != -1)
    {
        mouseUpRectangleSelection(e);
    }
}

void cEnvelopeComponent::mouseDrag(const MouseEvent &e)
{
    // if we are currently dragging an anchor
    if (mDraggingAnchor != kNoAnchor)
    {
        mouseDragAnchor(e);
    }
    else if (mStartRectSelectX != -1)
    {
        mouseDragRectangleSelection(e);
    }
}

void cEnvelopeComponent::mouseMove(const MouseEvent &e)
{
    // find the anchor which is under the mouse (or kNoAnchor)
    cEnvelopeAnchor* mouseOverAnchor = kNoAnchor;
    for (int curAnchor = 0; curAnchor < mEnvelopeAnchors.size(); ++curAnchor)
    {
        if (Rectangle<int>(mOffset + mEnvelopeAnchors[curAnchor]->mCurX - (mAnchorSize / 2), mOffset + mEnvelopeAnchors[curAnchor]->mCurY - (mAnchorSize / 2), mAnchorSize, mAnchorSize).contains(e.x, e.y))
        {
            mouseOverAnchor = mEnvelopeAnchors[curAnchor];
            break;
        }
    }

    // if the anchor has changed
    if (mouseOverAnchor != mCurActiveAnchor)
    {
        if (mCurActiveAnchor != kNoAnchor)
            mCurActiveAnchor->SetActive(false);

        mCurActiveAnchor = mouseOverAnchor;

        if (mCurActiveAnchor != kNoAnchor)
            mCurActiveAnchor->SetActive(true);

        repaint();
    }
}

void cEnvelopeComponent::resized()
{
    // calculate the actual display width, which includes subtracting 
    // the border width on either side
    mEnvelopeDisplayAreaWidth  = getWidth() - (mBorderWidth * 2) - mAnchorSize;
    mEnvelopeDisplayAreaHeight = getHeight() - (mBorderWidth * 2) - mAnchorSize;

    for( int curAnchor = 0; curAnchor < mEnvelopeAnchors.size(); ++curAnchor )
    {
        mEnvelopeAnchors[ curAnchor ]->SetMaxX(mEnvelopeDisplayAreaWidth);
        mEnvelopeAnchors[ curAnchor ]->SetMaxY(mEnvelopeDisplayAreaHeight);
    }

    recalculateCoordinates();
}

void cEnvelopeComponent::paintBackground( Graphics& g )
{
    // fill background with color
    //g.fillAll( mBackgroundColour );

//     g.setGradientFill (ColourGradient (Colours::black.withAlpha (0.3f), 0, 0,
//         Colours::transparentBlack, getWidth(), getHeight(), false));
//     g.fillRect(0, 0, getWidth(), getHeight());

    g.fillAll( Colours::white );
    g.drawImage( mBackgroundImage, 0, 0, getWidth(), getHeight(), 0, 0, mBackgroundImage.getWidth(), mBackgroundImage.getHeight(), false );

    // set color for border
    g.setColour( mBorderColour );

    // draw border
    g.drawRect( 0, 0, getWidth(), getHeight(), mBorderWidth );
}

void cEnvelopeComponent::paint(Graphics& g)
{
    if( !isEnabled() )
        g.setOpacity (0.5f);

    paintBackground( g );

    // draw envelope line between each two anchors (thus we only iterate thru size - 1)
    for( long curAnchor = 0; curAnchor < mEnvelopeAnchors.size() - 1; ++curAnchor )
    {
        Line<float> lineSegment((float)(mOffset + mEnvelopeAnchors[curAnchor]->mCurX), (float)(mOffset + mEnvelopeAnchors[curAnchor]->mCurY), (float)(mOffset + mEnvelopeAnchors[curAnchor + 1]->mCurX), (float)(mOffset + mEnvelopeAnchors[curAnchor + 1]->mCurY));

        // set color for the envelope
        g.setColour(mGlowColour.brighter(0.8f).withAlpha(0.4f));
        g.drawLine(lineSegment, 3.0f);

        // set color for the envelope
        g.setColour(mLineColour.brighter(0.2f));
        g.drawLine(lineSegment, 1.0f);
    }

    // draw anchors
    for (int curAnchor = 0; curAnchor < mEnvelopeAnchors.size(); ++curAnchor)
        mEnvelopeAnchors[ curAnchor ]->Display( g, mOffset, mOffset );

    // draw rectangular selection
    if (mStartRectSelectX > -1)
    {
        g.setColour(Colours::darkcyan);
        int width = mCurRectSelectX - mStartRectSelectX;
        int height = mCurRectSelectY - mStartRectSelectY;
        int startSelectX = mStartRectSelectX;
        int startSelectY = mStartRectSelectY;
        if (width < 0)
        {
            startSelectX += width;
            width = abs(width);
        }
        if (height < 0)
        {
            startSelectY += height;
            height = abs(height);
        }
        g.drawRect((float)startSelectX, (float)startSelectY, (float)width, (float)height, 1.0f);
    }
}
